@@ -1,13 +1,13 @@
 /* -*- tab-width:4;c-file-style:"cc-mode"; -*- */
 /*
  * avinfo.c -- Convert ffmpeg supported a/v files to  Ogg Theora / Vorbis
- * Copyright (C) 2003-2009 <j@v2v.cc>
+ * Copyright (C) 2003-2011 <j@v2v.cc>
  *
  *   gcc -o avinfo avinfo.c -DAVINFO `pkg-config --cflags --libs libavcodec libavformat`
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation, either version 3 of the License, or
+ * the Free Software Foundation, either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -42,6 +42,7 @@
 #include <sys/stat.h>
 
 #include "libavformat/avformat.h"
+#include "libavutil/pixdesc.h"
 
 #ifndef WIN32
 #if !defined(off64_t)
@@ -95,6 +96,9 @@ char const *fix_codec_name(char const *codec_name) {
     else if (!strcmp(codec_name, "0x0000")) {
         codec_name = "mu-law";
     }
+    else if (!strcmp(codec_name, "libvpx")) {
+        codec_name = "vp8";
+    }
    return codec_name;
 }
 
@@ -193,11 +197,11 @@ void json_codec_info(FILE *output, AVCodecContext *enc, int indent) {
     }
 
     switch(enc->codec_type) {
-    case CODEC_TYPE_VIDEO:
+    case AVMEDIA_TYPE_VIDEO:
         codec_name = fix_codec_name(codec_name);
         json_add_key_value(output, "codec", (void *)codec_name, JSON_STRING, 0, indent);
         if (enc->pix_fmt != PIX_FMT_NONE) {
-            json_add_key_value(output, "pixel_format", (void *)avcodec_get_pix_fmt_name(enc->pix_fmt), JSON_STRING, 0, indent);
+            json_add_key_value(output, "pixel_format", (void *)av_get_pix_fmt_name(enc->pix_fmt), JSON_STRING, 0, indent);
         }
         if (enc->width) {
             json_add_key_value(output, "width", &enc->width, JSON_INT, 0, indent);
@@ -221,7 +225,7 @@ void json_codec_info(FILE *output, AVCodecContext *enc, int indent) {
             json_add_key_value(output, "bitrate", &t, JSON_FLOAT, 0, indent);
         }
         break;
-    case CODEC_TYPE_AUDIO:
+    case AVMEDIA_TYPE_AUDIO:
         codec_name = fix_codec_name(codec_name);
         json_add_key_value(output, "codec", (void *)codec_name, JSON_STRING, 0, indent);
         if (enc->sample_rate) {
@@ -274,15 +278,15 @@ void json_codec_info(FILE *output, AVCodecContext *enc, int indent) {
         }
         break;
     /*
-    case CODEC_TYPE_DATA:
+    case AVMEDIA_TYPE_DATA:
         fprintf(output, "datacodec: %s\n", codec_name);
         bitrate = enc->bit_rate;
         break;
-    case CODEC_TYPE_SUBTITLE:
+    case AVMEDIA_TYPE_SUBTITLE:
         fprintf(output, "subtitle: %s\n", codec_name);
         bitrate = enc->bit_rate;
         break;
-    case CODEC_TYPE_ATTACHMENT:
+    case AVMEDIA_TYPE_ATTACHMENT:
         fprintf(output, "attachment: : %s\n", codec_name);
         bitrate = enc->bit_rate;
         break;
@@ -313,21 +317,7 @@ static void json_stream_format(FILE *output, AVFormatContext *ic, int i, int ind
         fprintf(output, "{\n");
 
         json_codec_info(output, st->codec, indent + 1);
-        if (st->sample_aspect_ratio.num && // default
-            av_cmp_q(st->sample_aspect_ratio, st->codec->sample_aspect_ratio)) {
-            AVRational display_aspect_ratio;
-            av_reduce(&display_aspect_ratio.num, &display_aspect_ratio.den,
-                      st->codec->width*st->sample_aspect_ratio.num,
-                      st->codec->height*st->sample_aspect_ratio.den,
-                      1024*1024);
-            snprintf(buf1, sizeof(buf1), "%d:%d",
-                     st->sample_aspect_ratio.num, st->sample_aspect_ratio.den);
-            json_add_key_value(output, "pixel_aspect_ratio", buf1, JSON_STRING, 0, indent + 1);
-            snprintf(buf1, sizeof(buf1), "%d:%d",
-                     display_aspect_ratio.num, display_aspect_ratio.den);
-            json_add_key_value(output, "display_aspect_ratio", buf1, JSON_STRING, 0, indent + 1);
-        }
-        if(st->codec->codec_type == CODEC_TYPE_VIDEO){
+        if(st->codec->codec_type == AVMEDIA_TYPE_VIDEO){
             if (st->time_base.den && st->time_base.num && av_q2d(st->time_base) > 0.001) {
                 snprintf(buf1, sizeof(buf1), "%d:%d",
                          st->time_base.den, st->time_base.num);
@@ -337,6 +327,20 @@ static void json_stream_format(FILE *output, AVFormatContext *ic, int i, int ind
                          st->r_frame_rate.num, st->r_frame_rate.den);
                 json_add_key_value(output, "framerate", buf1, JSON_STRING, 0, indent + 1);
             }
+            if (st->sample_aspect_ratio.num && // default
+                av_cmp_q(st->sample_aspect_ratio, st->codec->sample_aspect_ratio)) {
+                AVRational display_aspect_ratio;
+                av_reduce(&display_aspect_ratio.num, &display_aspect_ratio.den,
+                          st->codec->width*st->sample_aspect_ratio.num,
+                          st->codec->height*st->sample_aspect_ratio.den,
+                          1024*1024);
+                snprintf(buf1, sizeof(buf1), "%d:%d",
+                         st->sample_aspect_ratio.num, st->sample_aspect_ratio.den);
+                json_add_key_value(output, "pixel_aspect_ratio", buf1, JSON_STRING, 0, indent+1);
+                snprintf(buf1, sizeof(buf1), "%d:%d",
+                         display_aspect_ratio.num, display_aspect_ratio.den);
+                json_add_key_value(output, "display_aspect_ratio", buf1, JSON_STRING, 0, indent+1);
+            }
         }
         json_add_key_value(output, "id", &i, JSON_INT, 1, indent + 1);
         do_indent(output, indent-1);
@@ -355,6 +359,7 @@ static int utf8_validate (char *s, int n) {
       i+=4;
       continue;
     }
+    if (s[i] < 0) goto error;
     if (s[i] < 128) {
       i++;
       continue;
@@ -382,12 +387,11 @@ static int utf8_validate (char *s, int n) {
 error:
   return i == n;
 }
-
-void json_metadata(FILE *output, const AVFormatContext *av)
+int _json_metadata(FILE *output, AVDictionary *m, int first, int indent)
 {
-    int first = 1, indent=2;
-    AVMetadataTag *tag = NULL;
-    while ((tag = av_metadata_get(av->metadata, "", tag, AV_METADATA_IGNORE_SUFFIX))) {
+    int i = 0;
+    AVDictionaryEntry *tag = NULL;
+    while ((tag = av_dict_get(m, "", tag, AV_DICT_IGNORE_SUFFIX))) {
         if (strlen(tag->value) && utf8_validate (tag->value, strlen(tag->value))) {
             if (first) {
                 first = 0;
@@ -401,6 +405,17 @@ void json_metadata(FILE *output, const AVFormatContext *av)
             json_add_key_value(output, tag->key, tag->value, JSON_STRING, 1, indent);
         }
     }
+    return first;
+}
+
+void json_metadata(FILE *output, const AVFormatContext *av)
+{
+    int first = 1, indent=2, i=0;
+    first = _json_metadata(output, av->metadata, first, indent);
+
+    for(i=0;i<av->nb_streams;i++) {
+        first = _json_metadata(output, av->streams[i]->metadata, first, indent);
+    }
     if (!first) {
         do_indent(output, 1);
         fprintf(output, "},\n");
@@ -481,11 +496,11 @@ void json_format_info(FILE* output, AVFormatContext *ic, const char *url) {
             int j, k;
             for(j=0; j<ic->nb_programs; j++) {
                 for(k=0; k<ic->programs[j]->nb_stream_indexes; k++)
-                    json_stream_format(output, ic, ic->programs[j]->stream_index[k], 2, !k && !j, CODEC_TYPE_VIDEO);
+                    json_stream_format(output, ic, ic->programs[j]->stream_index[k], 2, !k && !j, AVMEDIA_TYPE_VIDEO);
              }
         } else {
             for(i=0;i<ic->nb_streams;i++) {
-                json_stream_format(output, ic, i, 2, !i, CODEC_TYPE_VIDEO);
+                json_stream_format(output, ic, i, 2, !i, AVMEDIA_TYPE_VIDEO);
             }
         }
         fprintf(output, "],\n");
@@ -496,11 +511,11 @@ void json_format_info(FILE* output, AVFormatContext *ic, const char *url) {
             int j, k;
             for(j=0; j<ic->nb_programs; j++) {
                 for(k=0; k<ic->programs[j]->nb_stream_indexes; k++)
-                    json_stream_format(output, ic, ic->programs[j]->stream_index[k], 2, !k && !j, CODEC_TYPE_AUDIO);
+                    json_stream_format(output, ic, ic->programs[j]->stream_index[k], 2, !k && !j, AVMEDIA_TYPE_AUDIO);
              }
         } else {
             for(i=0;i<ic->nb_streams;i++) {
-                json_stream_format(output, ic, i, 2, !i, CODEC_TYPE_AUDIO);
+                json_stream_format(output, ic, i, 2, !i, AVMEDIA_TYPE_AUDIO);
             }
         }
         fprintf(output, "],\n");
