@@ -1,11 +1,11 @@
 /* -*- tab-width:4;c-file-style:"cc-mode"; -*- */
 /*
  * theorautils.c - Ogg Theora/Ogg Vorbis Abstraction and Muxing
- * Copyright (C) 2003-2009 <j@v2v.cc>
+ * Copyright (C) 2003-2011 <j@v2v.cc>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation, either version 3 of the License, or
+ * the Free Software Foundation, either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -36,11 +36,6 @@
 #endif
 #endif
 
-#ifdef WIN32
-#define fseeko fseeko64
-#define ftello ftello64
-#endif
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -51,6 +46,13 @@
 #include <math.h>
 #include <limits.h>
 
+#ifdef WIN32
+#if !defined(fseeko)
+#define fseeko fseeko64
+#define ftello ftello64
+#endif
+#endif
+
 #include "theora/theoraenc.h"
 #include "vorbis/codec.h"
 #include "vorbis/vorbisenc.h"
@@ -64,7 +66,7 @@
 void init_info(oggmux_info *info) {
     info->output_seekable = MAYBE_SEEKABLE;
     info->with_skeleton = 1; /* skeleton is enabled by default    */
-    info->with_seek_index = 0; /* keyframe index disabled by default. */
+    info->skeleton_3 = 0; /* by default, output skeleton 4 with keyframe indexes. */
     info->index_interval = 2000;
     info->theora_index_reserve = -1;
     info->vorbis_index_reserve = -1;
@@ -106,6 +108,8 @@ void init_info(oggmux_info *info) {
 
     info->prev_vorbis_window = -1;
     info->content_offset = 0;
+
+    info->serialno = 0;
 }
 
 void oggmux_setup_kate_streams(oggmux_info *info, int n_kate_streams)
@@ -126,14 +130,6 @@ void oggmux_setup_kate_streams(oggmux_info *info, int n_kate_streams)
     }
 }
 
-static ogg_uint32_t read_uint32(unsigned const char* p) {
-  ogg_uint32_t i = p[0] +
-                  (p[1] << 8) + 
-                  (p[2] << 16) +
-                  (p[3] << 24);
-  return i;  
-}
-
 static void write16le(unsigned char *ptr,ogg_uint16_t v)
 {
     ptr[0]=v&0xff;
@@ -200,48 +196,10 @@ write_page(oggmux_info* info, ogg_page* page)
     assert(info->output_seekable != MAYBE_SEEKABLE);
 }
 
-#define MAX(a,b) (((a) > (b)) ? (a) : (b))
-#define MIN(a,b) (((a) < (b)) ? (a) : (b))
-
-static ogg_int64_t index_start_time(oggmux_info* info)
-{
-    ogg_int64_t start_time;
-    int n;
-
-    if (!info->with_seek_index || !info->indexing_complete) {
-        return -1;
-    }
-
-    start_time = MIN(info->theora_index.start_time, info->vorbis_index.start_time);
-    for (n=0; n<info->n_kate_streams; ++n) {
-        oggmux_kate_stream *ks=info->kate_streams+n;
-        if (ks->index.start_time < start_time)
-            start_time = ks->index.start_time;
-    }
-    return start_time;
-}
-
-static ogg_int64_t index_end_time(oggmux_info* info)
-{
-    ogg_int64_t end_time;
-    int n;
-
-    if (!info->with_seek_index || !info->indexing_complete) {
-        return -1;
-    }
-    end_time = MAX(info->theora_index.end_time, info->vorbis_index.end_time);
-    for (n=0; n<info->n_kate_streams; ++n) {
-        oggmux_kate_stream *ks=info->kate_streams+n;
-        if (ks->index.end_time > end_time)
-            end_time = ks->index.end_time;
-    }
-    return end_time;
-}
-
 static ogg_int64_t output_file_length(oggmux_info* info)
 {
     ogg_int64_t offset, length;
-    if (!info->with_seek_index || !info->indexing_complete) {
+    if (info->skeleton_3 || !info->indexing_complete) {
         return -1;
     }
     offset = ftello(info->outfile);
@@ -267,19 +225,9 @@ void add_fishead_packet (oggmux_info *info,
     ogg_uint32_t version = SKELETON_VERSION(ver_maj, ver_min);
 
     assert(version >= SKELETON_VERSION(3,0) ||
-           version <= SKELETON_VERSION(3,3));
+           version == SKELETON_VERSION(4,0));
 
-    switch (version) {
-        case SKELETON_VERSION(3,0):
-            packet_size = 64;
-            break;
-        case SKELETON_VERSION(3,3):
-            packet_size = 112;
-            break;
-        default:
-            fprintf (stderr, "ERROR: Unknown skeleton version\n");
-            exit (1);
-    };
+    packet_size = (version == SKELETON_VERSION(4,0)) ? 80 : 64;
 
     memset (&op, 0, sizeof (op));
 
@@ -298,13 +246,9 @@ void add_fishead_packet (oggmux_info *info,
     write32le(op.packet+44, 0); /* UTC time, set to zero for now */
 
     /* Index start/end time, if unknown or non-indexed, will be -1. */
-    if (version >= SKELETON_VERSION(3,2)) {
-        write64le(op.packet+64, index_start_time(info));
-        write64le(op.packet+72, (ogg_int64_t)1000);
-        write64le(op.packet+80, index_end_time(info));
-        write64le(op.packet+88, (ogg_int64_t)1000);
-        write64le(op.packet+96, output_file_length(info));
-        write64le(op.packet+104, info->content_offset);
+    if (version == SKELETON_VERSION(4,0)) {
+        write64le(op.packet+64, output_file_length(info));
+        write64le(op.packet+72, info->content_offset);
     }
     op.b_o_s = 1; /* its the first packet of the stream */
     op.e_o_s = 0; /* its not the last packet of the stream */
@@ -314,18 +258,32 @@ void add_fishead_packet (oggmux_info *info,
     _ogg_free (op.packet);
 }
 
+const char* theora_message_headers = "Content-Type: video/theora\r\n"
+                                     "Role: video/main\r\n"
+                                     "Name: video_1\r\n";
+
+const char* vorbis_message_headers = "Content-Type: audio/vorbis\r\n"
+                                     "Role: audio/main\r\n"
+                                     "Name: audio_1\r\n";
+#ifdef HAVE_KATE
+const char* kate_message_headers =   "Content-Type: application/x-kate\r\n\r\n"
+                                     "Role: text/subtitle\r\n";
+                                     /* Dynamically add name header... */
+#endif
+
 /*
  * Adds the fishead packets in the skeleton output stream along with the e_o_s packet
  */
 void add_fisbone_packet (oggmux_info *info) {
     ogg_packet op;
-
+    size_t packet_size = 0;
     if (!info->audio_only) {
         memset (&op, 0, sizeof (op));
-        op.packet = _ogg_calloc (80, sizeof(unsigned char));
+        packet_size = FISBONE_SIZE + strlen(theora_message_headers);
+        op.packet = _ogg_calloc (packet_size, sizeof(unsigned char));
         if (op.packet == NULL) return;
 
-        memset (op.packet, 0, 80);
+        memset (op.packet, 0, packet_size);
         /* it will be the fisbone packet for the theora video */
         memcpy (op.packet, FISBONE_IDENTIFIER, 8); /* identifier */
         write32le(op.packet+8, FISBONE_MESSAGE_HEADER_OFFSET); /* offset of the message header fields */
@@ -337,11 +295,12 @@ void add_fisbone_packet (oggmux_info *info) {
         write64le(op.packet+36, 0); /* start granule */
         write32le(op.packet+44, 0); /* preroll, for theora its 0 */
         *(op.packet+48) = info->ti.keyframe_granule_shift; /* granule shift */
-        memcpy(op.packet+FISBONE_SIZE, "Content-Type: video/theora\r\n", 28); /* message header field, Content-Type */
+        /* message header fields */
+        memcpy(op.packet+FISBONE_SIZE, theora_message_headers, strlen(theora_message_headers));
 
         op.b_o_s = 0;
         op.e_o_s = 0;
-        op.bytes = 80; /* size of the packet in bytes */
+        op.bytes = packet_size; /* size of the packet in bytes */
 
         ogg_stream_packetin (&info->so, &op);
         _ogg_free (op.packet);
@@ -349,10 +308,11 @@ void add_fisbone_packet (oggmux_info *info) {
 
     if (!info->video_only) {
         memset (&op, 0, sizeof (op));
-        op.packet = _ogg_calloc (80, sizeof(unsigned char));
+        packet_size = FISBONE_SIZE + strlen(vorbis_message_headers);
+        op.packet = _ogg_calloc (packet_size, sizeof(unsigned char));
         if (op.packet == NULL) return;
 
-        memset (op.packet, 0, 80);
+        memset (op.packet, 0, packet_size);
         /* it will be the fisbone packet for the vorbis audio */
         memcpy (op.packet, FISBONE_IDENTIFIER, 8); /* identifier */
         write32le(op.packet+8, FISBONE_MESSAGE_HEADER_OFFSET); /* offset of the message header fields */
@@ -364,12 +324,13 @@ void add_fisbone_packet (oggmux_info *info) {
         write64le(op.packet+36, 0); /* start granule */
         write32le(op.packet+44, 2); /* preroll, for vorbis its 2 */
         *(op.packet+48) = 0; /* granule shift, always 0 for vorbis */
-        memcpy (op.packet+FISBONE_SIZE, "Content-Type: audio/vorbis\r\n", 28);
+        memcpy(op.packet+FISBONE_SIZE, vorbis_message_headers, strlen(vorbis_message_headers));
+
         /* Important: Check the case of Content-Type for correctness */
 
         op.b_o_s = 0;
         op.e_o_s = 0;
-        op.bytes = 80;
+        op.bytes = packet_size;
 
         ogg_stream_packetin (&info->so, &op);
         _ogg_free (op.packet);
@@ -378,31 +339,39 @@ void add_fisbone_packet (oggmux_info *info) {
 #ifdef HAVE_KATE
     if (info->with_kate) {
         int n;
+        char name[32];
         for (n=0; n<info->n_kate_streams; ++n) {
+            size_t packet_size = 0;
+            int message_headers_len = strlen(kate_message_headers);
+            int name_len = 0;
             oggmux_kate_stream *ks=info->kate_streams+n;
-        memset (&op, 0, sizeof (op));
-        op.packet = _ogg_calloc (86, sizeof(unsigned char));
-        memset (op.packet, 0, 86);
+            memset (&op, 0, sizeof (op));
+            sprintf(name, "Name: %d\r\n", (n+1));
+            name_len = strlen(name);
+            packet_size = FISBONE_SIZE + message_headers_len + name_len;
+            op.packet = _ogg_calloc (packet_size, sizeof(unsigned char));
+            memset (op.packet, 0, packet_size);
             /* it will be the fisbone packet for the kate stream */
-        memcpy (op.packet, FISBONE_IDENTIFIER, 8); /* identifier */
+            memcpy (op.packet, FISBONE_IDENTIFIER, 8); /* identifier */
             write32le(op.packet+8, FISBONE_MESSAGE_HEADER_OFFSET); /* offset of the message header fields */
-        write32le(op.packet+12, ks->ko.serialno); /* serialno of the vorbis stream */
+            write32le(op.packet+12, ks->ko.serialno); /* serialno of the vorbis stream */
             write32le(op.packet+16, ks->ki.num_headers); /* number of header packet */
-        /* granulerate, temporal resolution of the bitstream in Hz */
-        write64le(op.packet+20, ks->ki.gps_numerator); /* granulerate numerator */
+            /* granulerate, temporal resolution of the bitstream in Hz */
+            write64le(op.packet+20, ks->ki.gps_numerator); /* granulerate numerator */
             write64le(op.packet+28, ks->ki.gps_denominator); /* granulerate denominator */
-        write64le(op.packet+36, 0); /* start granule */
+            write64le(op.packet+36, 0); /* start granule */
             write32le(op.packet+44, 0); /* preroll, for kate it's 0 */
-        *(op.packet+48) = ks->ki.granule_shift; /* granule shift */
-            memcpy (op.packet+FISBONE_SIZE, "Content-Type: application/x-kate\r\n", 34);
-        /* Important: Check the case of Content-Type for correctness */
+            *(op.packet+48) = ks->ki.granule_shift; /* granule shift */
+            memcpy (op.packet+FISBONE_SIZE, kate_message_headers, message_headers_len);
+            memcpy (op.packet+FISBONE_SIZE+message_headers_len, name, name_len);
+            /* Important: Check the case of Content-Type for correctness */
 
-        op.b_o_s = 0;
-        op.e_o_s = 0;
-        op.bytes = 86;
+            op.b_o_s = 0;
+            op.e_o_s = 0;
+            op.bytes = packet_size;
 
             ogg_stream_packetin (&info->so, &op);
-        _ogg_free (op.packet);
+            _ogg_free (op.packet);
         }
     }
 #endif
@@ -420,7 +389,7 @@ static int create_index_packet(size_t bytes,
                                ogg_uint32_t serialno,
                                ogg_int64_t num_used_keypoints)
 {
-    size_t size = 26 + bytes;
+    size_t size = 42 + bytes;
     memset (op, 0, sizeof(*op));
     op->packet = malloc(size);
     if (op->packet == NULL)
@@ -442,6 +411,12 @@ static int create_index_packet(size_t bytes,
     /* Write timestamp denominator, times are in milliseconds, so 1000. */
     write64le(op->packet+18, (ogg_int64_t)1000);
 
+    /* Write first sample time numerator. */
+    write64le(op->packet+26, (ogg_int64_t)0);
+
+    /* Write last sample time numerator. */
+    write64le(op->packet+34, (ogg_int64_t)0);
+
     return 0;
 }
 
@@ -560,8 +535,12 @@ write_index_pages (seek_index* index,
     int index_bytes = 0;  
     int keypoints_cutoff = 0;
 
-    /* Must have indexed keypoints. */
-    assert(index->max_keypoints > 0 && index->packet_num > 0);
+    /* Must have indexed keypoints to go on */
+    if (index->max_keypoints == 0 || index->packet_num == 0) {
+      fprintf(stderr, "WARNING: no key points for %s stream %08x\n", name, serialno);
+      return 0;
+    }
+
     /* Must have placeholder packet to rewrite. */
     assert(index->page_location > 0);
 
@@ -628,11 +607,16 @@ write_index_pages (seek_index* index,
                "only part of the file may be indexed. Rerun with --%s-index-reserve %d to "
                "ensure a complete index, or use OggIndex to re-index.\n",
                name, (k - keypoints_cutoff), name, index_bytes);
-    } else if (index_bytes < index->packet_size) {
+    } else if (index_bytes < index->packet_size &&
+               index->packet_size - index_bytes > 10000)
+    {
+        /* We over estimated the index size by 10,000 bytes or more. */
         printf("Allocated %d bytes for %s keyframe index, %d are unused. "
+               "Index contains %d keyframes. "
                "Rerun with '--%s-index-reserve %d' to encode with the optimal sized %s index,"
                " or use OggIndex to re-index.\n",
                index->packet_size, name, (index->packet_size - index_bytes),
+               keypoints_cutoff,
                name, index_bytes, name);
     }
     num_keypoints = keypoints_cutoff;
@@ -645,9 +629,15 @@ write_index_pages (seek_index* index,
         free(keypoints);
         return -1;
     }
+
+    /* Write first sample time numerator. */
+    write64le(op.packet+26, index->start_time);
+
+    /* Write last sample time numerator. */
+    write64le(op.packet+34, index->end_time);
    
     /* Write keypoint data into packet. */
-    p = op.packet + 26;
+    p = op.packet + 42;
     limit = op.packet + op.bytes;
     prev_offset = 0;
     prev_time = 0;
@@ -697,8 +687,8 @@ int write_seek_index (oggmux_info* info)
     ogg_uint32_t serialno;
     ogg_page og;
 
-    /* If the index is disabled, we shouldn't be doing this! */
-    assert(info->with_seek_index);
+    /* We shouldn't write indexes for skeleton 3, it's a skeleton 4 feature. */
+    assert(!info->skeleton_3);
 
     /* Mark that we're done indexing. This causes the header packets' fields
        to be filled with valid, non-unknown values. */
@@ -710,7 +700,7 @@ int write_seek_index (oggmux_info* info)
     ogg_stream_clear(&info->so);
     ogg_stream_init(&info->so, serialno);
 
-    add_fishead_packet (info, 3, 3);
+    add_fishead_packet (info, 4, 0);
     if (ogg_stream_flush(&info->so, &og) != 1) {
         fprintf (stderr, "Internal Ogg library error.\n");
         exit (1);
@@ -831,7 +821,8 @@ void oggmux_init (oggmux_info *info) {
 
     /* yayness.  Set up Ogg output stream */
     srand (time (NULL));
-    ogg_stream_init (&info->vo, rand ());
+    info->serialno = rand();
+    ogg_stream_init (&info->vo, info->serialno++);
 
     if (info->passno!=1) {
         th_comment_add_tag(&info->tc, "ENCODER", PACKAGE_STRING);
@@ -843,7 +834,7 @@ void oggmux_init (oggmux_info *info) {
     }
 
     if (!info->audio_only) {
-        ogg_stream_init (&info->to, rand ());    /* oops, add one ot the above */
+        ogg_stream_init (&info->to, info->serialno++);
         seek_index_init(&info->theora_index, info->index_interval);
     }
     /* init theora done */
@@ -869,6 +860,7 @@ void oggmux_init (oggmux_info *info) {
         vorbis_block_init (&info->vd, &info->vb);
         
         seek_index_init(&info->vorbis_index, info->index_interval);
+        info->vorbis_granulepos = 0;
     }
     /* audio init done */
 
@@ -878,7 +870,7 @@ void oggmux_init (oggmux_info *info) {
         int ret, n;
         for (n=0; n<info->n_kate_streams; ++n) {
             oggmux_kate_stream *ks=info->kate_streams+n;
-            ogg_stream_init (&ks->ko, rand ());    /* oops, add one ot the above */
+            ogg_stream_init (&ks->ko, info->serialno++);
             ret = kate_encode_init (&ks->k, &ks->ki);
             if (ret<0) {
                 fprintf(stderr, "kate_encode_init: %d\n",ret);
@@ -897,27 +889,26 @@ void oggmux_init (oggmux_info *info) {
     }
     /* kate init done */
 
-    if (info->with_seek_index &&
+    if (info->with_skeleton &&
+        !info->skeleton_3 &&
         info->duration == -1)
     {
         /* We've not got a duration, we can't index the keyframes. */
-        fprintf(stderr, "WARNING: Can't get duration of media, not indexing.\n");
-        info->with_seek_index = 0;
+        fprintf(stderr, "WARNING: Can't get duration of media, not indexing, writing Skeleton 3 track.\n");
+        info->skeleton_3 = 1;
     }
-    /* We must only index keyframes when the skeleton track is also present. */
-    assert (!info->with_seek_index || info->with_skeleton);
 
     /* first packet should be skeleton fishead packet, if skeleton is used */
 
     if (info->with_skeleton && info->passno!=1) {
         /* Sometimes the output file is not seekable. We can't write the seek
            index if the output is not seekable. So write a Skeleton3.0 header
-           packet, which will in turn determines if the file is seekable. If it
+           packet, which will in turn determine if the file is seekable. If it
            is, we can safely construct an index, so then overwrite the header
-           page with a Skeleton3.1 header page. */
-        int with_seek_index = info->with_seek_index;
-        info->with_seek_index = 0;
-        ogg_stream_init (&info->so, rand());
+           page with a Skeleton4.0 header page. */
+        int skeleton_3 = info->skeleton_3;
+        info->skeleton_3 = 1;
+        ogg_stream_init (&info->so, info->serialno++);
         add_fishead_packet (info, 3, 0);
         if (ogg_stream_pageout (&info->so, &og) != 1) {
             fprintf (stderr, "Internal Ogg library error.\n");
@@ -926,24 +917,23 @@ void oggmux_init (oggmux_info *info) {
         write_page (info, &og);
         assert(info->output_seekable != MAYBE_SEEKABLE);
 
-        if (info->output_seekable == NOT_SEEKABLE && with_seek_index) {
+        if (info->output_seekable == NOT_SEEKABLE && !skeleton_3) {
             fprintf(stderr, "WARNING: Can't write keyframe-seek-index into "
-                            "non-seekable output stream!\n");
+                            "non-seekable output stream! Writing Skeleton3 track.\n");
         }
 
-        info->with_seek_index = with_seek_index &&
-                                info->output_seekable == SEEKABLE;
+        info->skeleton_3 = skeleton_3 || info->output_seekable == NOT_SEEKABLE;
 
-        if (info->with_seek_index) {
+        if (!info->skeleton_3) {
             /* Output is seekable and we're indexing. Overwrite the
-               Skeleton3.0 BOS page with a Skeleton3.1 BOS page. */
+               Skeleton3.0 BOS page with a Skeleton4.0 BOS page. */
             if (fseeko (info->outfile, 0, SEEK_SET) < 0) {
                 fprintf (stderr, "ERROR: failed to seek in seekable output file!?!\n");
                 exit (1);
             }
             ogg_stream_clear (&info->so);
-            ogg_stream_init (&info->so, rand());
-            add_fishead_packet (info, 3, 3);
+            ogg_stream_init (&info->so, info->serialno++);
+            add_fishead_packet (info, 4, 0);
             if (ogg_stream_pageout (&info->so, &og) != 1) {
                 fprintf (stderr, "Internal Ogg library error.\n");
                 exit (1);
@@ -1094,7 +1084,7 @@ void oggmux_init (oggmux_info *info) {
     if (info->with_skeleton && info->passno!=1) {
         int result;
 
-        if (info->with_seek_index) {
+        if (!info->skeleton_3) {
             /* Add placeholder packets to reserve space for the index
              * at the start of file. */
             write_placeholder_index_pages (info);
@@ -1184,7 +1174,7 @@ void oggmux_add_video (oggmux_info *info, th_ycbcr_buffer ycbcr, int e_o_s) {
     }
 
     while (th_encode_packetout (info->td, e_o_s, &op) > 0) {
-        if (info->with_seek_index &&
+        if (!info->skeleton_3 &&
             info->passno != 1)
         {
             ogg_int64_t frameno = th_granule_frame(info->td, op.granulepos);
@@ -1236,7 +1226,7 @@ vorbis_time(vorbis_dsp_state * dsp, ogg_int64_t granulepos) {
 void oggmux_add_audio (oggmux_info *info, int16_t * buffer, int bytes, int samples, int e_o_s) {
     ogg_packet op;
 
-    int i,j, count = 0;
+    int i, j, k, count = 0;
     float **vorbis_buffer;
 
     if (bytes <= 0 && samples <= 0) {
@@ -1249,7 +1239,20 @@ void oggmux_add_audio (oggmux_info *info, int16_t * buffer, int bytes, int sampl
         /* uninterleave samples */
         for (i = 0; i < samples; i++) {
             for (j=0;j<info->channels;j++) {
-                vorbis_buffer[j][i] = buffer[count++] / 32768.f;
+                k = j;
+                /* 5.1 input: [fl, fr, c, lfe, rl, rr] */
+                if(info->channels == 6) {
+                    switch(j) {
+                        case 0: k = 0; break;
+                        case 1: k = 2; break;
+                        case 2: k = 1; break;
+                        case 3: k = 5; break;
+                        case 4: k = 3; break;
+                        case 5: k = 4; break;
+                        default: k = j;
+                    }
+                }
+                vorbis_buffer[k][i] = buffer[count++] / 32768.f;
             }
         }
         vorbis_analysis_wrote (&info->vd, samples);
@@ -1280,10 +1283,35 @@ void oggmux_add_audio (oggmux_info *info, int16_t * buffer, int bytes, int sampl
             info->prev_vorbis_window = info->vb.pcmend;
 
             ogg_int64_t start_granule = op.granulepos - num_samples;
+            if (start_granule < 0) {
+                /* The first vorbis content packet can have more samples than
+                   its granulepos reports. This is allowed by the spec, and
+                   players should discard the leading samples and not play them.
+                   Thus the indexer needs to discard them as well.*/
+                if (op.packetno != 4) {
+                    /* We only expect negative start granule in the first content
+                       packet, not any of the others... */
+                    fprintf(stderr, "WARNING: vorbis packet %lld has calculated start"
+                            " granule of %lld, but it should be non-negative!",
+                            op.packetno, start_granule);
+                }
+                start_granule = 0;
+            }
+            if (start_granule < info->vorbis_granulepos) {
+                /* This packet starts before the end of the previous packet. This is
+                   allowed by the specification in the last packet only, and the
+                   trailing samples should be discarded and not played/indexed. */
+                if (!op.e_o_s) {
+                    fprintf(stderr, "WARNING: vorbis packet %lld (granulepos %lld) starts before"
+                            " the end of the preceeding packet!", op.packetno, op.granulepos);
+                }
+                start_granule = info->vorbis_granulepos;
+            }
+            info->vorbis_granulepos = op.granulepos;
             ogg_int64_t start_time = vorbis_time (&info->vd, start_granule);
             
             if (op.granulepos != -1 &&
-                info->with_seek_index &&
+                !info->skeleton_3 &&
                 info->passno != 1)
             {
                 ogg_int64_t end_time = vorbis_time (&info->vd, op.granulepos);
@@ -1312,6 +1340,8 @@ static void oggmux_record_kate_index(oggmux_info *info, oggmux_kate_stream *ks,
     ks->last_end_time = end_time;
 }
 
+#ifdef HAVE_KATE
+
 /**
  * adds a subtitles text to the encoding sink
  * if e_o_s is 1 the end of the logical bitstream will be marked.
@@ -1322,14 +1352,27 @@ static void oggmux_record_kate_index(oggmux_info *info, oggmux_kate_stream *ks,
  * @param text the utf-8 text
  * @param len the number of bytes in the text
  */
-void oggmux_add_kate_text (oggmux_info *info, int idx, double t0, double t1, const char *text, size_t len) {
-#ifdef HAVE_KATE
+void oggmux_add_kate_text (oggmux_info *info, int idx, double t0, double t1, const char *text, size_t len, int x1, int x2, int y1, int y2) {
     ogg_packet op;
     oggmux_kate_stream *ks=info->kate_streams+idx;
     int ret;
+    if (x1!=-INT_MAX && y1!=-INT_MAX && x2!=-INT_MAX && y2!=-INT_MAX) {
+      kate_region kr;
+      kate_region_init(&kr);
+      kr.metric=kate_pixel;
+      kr.x=x1;
+      kr.y=y1;
+      kr.w=x2-x1+1;
+      kr.h=y2-y1+1;
+      ret=kate_encode_set_region(&ks->k,&kr);
+      if (ret<0) {
+        fprintf(stderr,"Error setting region: %d\n",ret);
+        return;
+      }
+    }
     ret = kate_ogg_encode_text(&ks->k, t0, t1, text, len, &op);
     if (ret>=0) {
-        if (info->with_seek_index && info->passno != 1) {
+        if (!info->skeleton_3 && info->passno != 1) {
             ogg_int64_t start_time = (int)(t0 * 1000.0f + 0.5f);
             ogg_int64_t end_time = (int)(t1 * 1000.0f + 0.5f);
             oggmux_record_kate_index(info, ks, &op, start_time, end_time);
@@ -1343,7 +1386,42 @@ void oggmux_add_kate_text (oggmux_info *info, int idx, double t0, double t1, con
         fprintf(stderr, "Failed to encode kate data packet (%f --> %f, [%s]): %d\n",
             t0, t1, text, ret);
     }
-#endif
+}
+
+/**
+ * adds a subtitles image to the encoding sink
+ * if e_o_s is 1 the end of the logical bitstream will be marked.
+ * @param info oggmux_info
+ * @param idx which kate stream to output to
+ * @param t0 the show time of the text
+ * @param t1 the hide time of the text
+ * @param kr the region in which to display the subtitle
+ * @param kp the palette to use for the image
+ * @param kb the image itself
+ */
+void oggmux_add_kate_image (oggmux_info *info, int idx, double t0, double t1, const kate_region *kr, const kate_palette *kp, const kate_bitmap *kb) {
+    ogg_packet op;
+    oggmux_kate_stream *ks=info->kate_streams+idx;
+    int ret;
+    ret = kate_encode_set_region(&ks->k, kr);
+    if (ret >= 0) ret = kate_encode_set_palette(&ks->k, kp);
+    if (ret >= 0) ret = kate_encode_set_bitmap(&ks->k, kb);
+    if (ret >= 0) ret = kate_ogg_encode_text(&ks->k, t0, t1, "", 0, &op);
+    if (ret>=0) {
+        if (!info->skeleton_3 && info->passno != 1) {
+            ogg_int64_t start_time = (int)(t0 * 1000.0f + 0.5f);
+            ogg_int64_t end_time = (int)(t1 * 1000.0f + 0.5f);
+            oggmux_record_kate_index(info, ks, &op, start_time, end_time);
+        }
+
+        ogg_stream_packetin (&ks->ko, &op);
+        ogg_packet_clear (&op);
+        info->k_pkg++;
+    }
+    else {
+        fprintf(stderr, "Failed to encode kate data packet (%f --> %f, image): %d\n",
+            t0, t1, ret);
+    }
 }
 
 /**
@@ -1353,13 +1431,12 @@ void oggmux_add_kate_text (oggmux_info *info, int idx, double t0, double t1, con
  * @param t the time of the end packet
  */
 void oggmux_add_kate_end_packet (oggmux_info *info, int idx, double t) {
-#ifdef HAVE_KATE
     ogg_packet op;
     oggmux_kate_stream *ks=info->kate_streams+idx;
     int ret;
     ret = kate_ogg_encode_finish(&ks->k, t, &op);
     if (ret>=0) {
-        if (info->with_seek_index && info->passno != 1) {
+        if (!info->skeleton_3 && info->passno != 1) {
             ogg_int64_t start_time = floorf(t * 1000.0f + 0.5f);
             ogg_int64_t end_time = start_time;
             oggmux_record_kate_index(info, ks, &op, start_time, end_time);
@@ -1371,9 +1448,10 @@ void oggmux_add_kate_end_packet (oggmux_info *info, int idx, double t) {
     else {
         fprintf(stderr, "Failed to encode kate end packet at %f: %d\n", t, ret);
     }
-#endif
 }
 
+#endif
+
 static double get_remaining(oggmux_info *info, double timebase) {
   double remaining = 0;
   double to_encode, time_so_far;
@@ -1430,12 +1508,21 @@ static void print_stats(oggmux_info *info, double timebase) {
     else if (timebase - last > 0.5 || timebase < last || !remaining) {
         last = timebase;
         if (info->frontend) {
+#ifdef WIN32
+            fprintf(info->frontend, "{\"duration\": %f, \"position\": %.02f, \"audio_kbps\":  %d, \"video_kbps\": %d, \"remaining\": %.02f}\n",
+                info->duration,
+                timebase,
+                info->akbps, info->vkbps,
+                remaining
+            );
+#else
             fprintf(info->frontend, "{\"duration\": %lf, \"position\": %.02lf, \"audio_kbps\":  %d, \"video_kbps\": %d, \"remaining\": %.02lf}\n",
                 info->duration,
                 timebase,
                 info->akbps, info->vkbps,
                 remaining
             );
+#endif
             fflush (info->frontend);
         }
         else if (timebase > 0) {
@@ -1485,7 +1572,6 @@ static void write_audio_page(oggmux_info *info)
     ogg_int64_t page_offset = ftello(info->outfile);
     int packets = ogg_page_packets((ogg_page *)&info->audiopage);
     int packet_start_num = ogg_page_start_packets(info->audiopage);
-    ogg_uint32_t checksum = read_uint32(info->audiopage + 22);
 
     ret = fwrite(info->audiopage, 1, info->audiopage_len, info->outfile);
     if (ret < info->audiopage_len) {
@@ -1519,7 +1605,6 @@ static void write_video_page(oggmux_info *info)
     ogg_int64_t page_offset = ftello(info->outfile);
     int packets = ogg_page_packets((ogg_page *)&info->videopage);
     int packet_start_num = ogg_page_start_packets(info->videopage);
-    ogg_uint32_t checksum = read_uint32(info->videopage + 22);
 
     ret = fwrite(info->videopage, 1, info->videopage_len, info->outfile);
     if (ret < info->videopage_len) {
